/*-----------------------------------------------------------------------

File  : ast.c

Author: Stephan Schulz (schulz@eprover.org)

Contents
 
  Implementation of an abstract syntax tree type.

  Copyright 2015 by the author.
  This code is released under the GNU General Public Licence.

Changes

<1> Tue Mar 31 18:39:18 CEST 2015
    New

-----------------------------------------------------------------------*/

#include "ast.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

long nodectr=0;

char* ast_name[] =
{
   "nil",
   "prog",
   "def",
   "vardef",
   "fundef",
   "type",
   "params",
   "paramlist",
   "param",
   "body",
   "vardefs",
   "stmts",
   "stmt",
   "while_stmt",
   "if_stmt",
   "ret_stmt",
   "print_stmt",
   "assign",
   "funcall_stmt",
   "boolexpr",
   "expr",
   "funcall",
   "args",
   "arglist",
   "arg",
   "t_MULT",
   "t_DIV",
   "t_PLUS",
   "t_MINUS",
   "t_EQ",
   "t_NEQ",
   "t_LT", 
   "t_GT", 
   "t_LEQ", 
   "t_GEQ",
   "t_IDENT",
   "t_STRINGLIT",
   "t_INTLIT",
   "t_INTEGER",
   "t_STRING",
   "t_IF",
   "t_WHILE",
   "t_RETURN",
   "t_PRINT",
   "t_dummy",
   "XXXX"
};
   
/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: ASTCellPrint()
//
//   Print an ASTCell to stdout (for debugging).
//
// Global Variables: Reads ast_name[]
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ASTCellPrint(AST_p cell)
{
   printf("ASTCell(%d, %s, %d, %d, %s, %ld, %p, %p, %p, %p)\n",
          cell->type, 
          ast_name[cell->type],
          cell->line, 
          cell->col, 
          cell->litval, 
          cell->intval, 
          cell->child1, 
          cell->child2, 
          cell->child3, 
          cell->child4); 
}


/*-----------------------------------------------------------------------
//
// Function: ASTAlloc()
//
//   Allocate a new ASTCell, initialize it with the given values, and
//   return a pointer to it.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

AST_p  ASTAlloc(ASTNodeType type, int line, int col, char* litval, 
                long intval, AST_p child1, AST_p child2, AST_p child3, 
                AST_p child4)
{
   AST_p ast = ASTCellAlloc();

   ast->type = type;
   ast->nodectr = nodectr++;
   ast->line = line;
   ast->col  = col;
   ast->litval = litval;
   ast->intval = intval;
   ast->child1 = child1;
   ast->child2 = child2;
   ast->child3 = child3;
   ast->child4 = child4;
   /* ASTCellPrint(ast); */

   return ast;
}


void   ASTFree(AST_p junk);


/*-----------------------------------------------------------------------
//
// Function: dot_print_child()
//
//   Print a link from father node to child node, then print the
//   subtree corresponding to the child in DOT syntax.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void dot_print_child(FILE* out, AST_p father, AST_p child)
{
   if(child)
   {
      fprintf(out, "   ast%ld -> ast%ld\n", father->nodectr, child->nodectr);
      DOTASTNodePrint(out, child);
   }
}

/*-----------------------------------------------------------------------
//
// Function: DOTASTNodePrint()
//
//   Print an AST node and its children in DOT notation.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void DOTASTNodePrint(FILE* out, AST_p ast)
{
   if(!ast)
   {
      return;
   }
   switch(ast->type)
   {
   case t_STRINGLIT:
         fprintf(out, "   ast%ld [label=%s]\n", 
                 ast->nodectr,
                 //ast_name[ast->type],
                 ast->litval);
         break;
   case t_IDENT:
   case t_INTLIT:
         fprintf(out, "   ast%ld [label=\"%s\\n%s\"]\n", 
                 ast->nodectr,
                 ast_name[ast->type],
                 ast->litval);
         break;
   default:
         fprintf(out, "   ast%ld [label=\"%s\"]\n", 
                 ast->nodectr,
                 ast_name[ast->type]);
         break;
   }
   dot_print_child(out, ast, ast->child1);
   dot_print_child(out, ast, ast->child2);
   dot_print_child(out, ast, ast->child3);
   dot_print_child(out, ast, ast->child4);
}

/*-----------------------------------------------------------------------
//
// Function: DOTASTPrint()
//
//   Print a complete AST as a DOT format graph.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void DOTASTPrint(FILE* out, AST_p ast)
{
   fprintf(out, "digraph ast {\n   ordering=out\n");
   DOTASTNodePrint(out, ast);
   fprintf(out, "}\n");
}



/*-----------------------------------------------------------------------
//
// Function: SExprASTPrint()
//
//   Print an AST representing a nanoLang construct as an
//   s-expression.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void SExprASTPrint(FILE* out, AST_p ast)
{
   if(ast)
   {
      switch(ast->type)
      {
      case t_STRINGLIT:
            fprintf(out, "t_STRINGLIT<%s> ",
                    ast->litval);
            break;
      case t_IDENT:
            fprintf(out, "t_IDENT<%s> ",
                    ast->litval);
            break;
      case t_INTLIT:
            fprintf(out, "t_INTLIT<%s> ",
                    ast->litval);
            break;
      default:
            fprintf(out, "(");
            fprintf(out, "%s ", ast_name[ast->type]);               
            SExprASTPrint(out, ast->child1);
            SExprASTPrint(out, ast->child2);
            SExprASTPrint(out, ast->child3);
            SExprASTPrint(out, ast->child4);      
            fprintf(out, ") ");         
            break;
      }      
   }
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


