 /* 
     nanoLang parser

     Copyright 2015 by Stephan Schulz, schulz@eprover.org.

     This code is released under the GNU General Public Licence
     Version 2.
 */

%{
   #include <stdio.h>
   #include "ast.h"
   
   extern int yyerror(const char* err);
   extern int yylex(void);
   extern FILE *yyin;
   
   AST_p ast = NULL;

%}

%locations
%define api.value.type {AST_p}

%define parse.error verbose

%start prog

%token IDENT
%token STRINGLIT
%token INTLIT
%token INTEGER STRING
%token IF WHILE RETURN PRINT
%token EQ NEQ LT GT LEQ GEQ
%left  PLUS MINUS
%left  MULT DIV
%right UMINUS
%token OPENPAR CLOSEPAR
%token SEMICOLON COMA
%token OPENCURLY CLOSECURLY


%token ERROR


%%

prog: /* Nothing */ {$$=NULL; ast=$$;}
    | prog def {
       if($1)
       {
          $$ = ASTAlloc2(prog, $1->line, $1->col, NULL, 0, $1, $2);
       }
       else
       {
          $$ = ASTAlloc2(prog, $2->line, $2->col, NULL, 0, $1, $2);
       }
       ast = $$;}
;

def: vardef {$$=$1;}
   | fundef {$$=$1;}
;

vardef: type IDENT SEMICOLON {
   $$ = ASTAlloc2(vardef, $1->line, $1->col, NULL, 0, $1, $2);}
;

fundef: type IDENT OPENPAR params CLOSEPAR body {
   $$ = ASTAlloc(fundef, $1->line, $1->col, NULL, 0, $1, $2, $4, $6);}

;

type: STRING  {$$ = $1;}
    | INTEGER {$$ = $1;}
;

params: /* empty */ {$$ = ASTAlloc2(params, 0,0,NULL,0,NULL,NULL);}
      | paramlist {$$ = ASTAlloc2(params, $1->line,$1->col,NULL,0,$1,NULL);}
;

paramlist: param {$$ = $1;}
         | paramlist COMA param {
            $$ = ASTAlloc2(paramlist, $1->line, $1->col, NULL, 0, $1, $3);}
;

param: type IDENT {$$ = ASTAlloc2(param, $1->line, $1->col, NULL, 0, $1, $2);}
;

body: OPENCURLY vardefs stmts CLOSECURLY {
   $$ = ASTAlloc2(body, $1->line, $1->col, NULL, 0, $2, $3);}
;

vardefs: /* empty */ {$$=NULL;}
       | vardefs vardef {
          if($1)
          {
             $$ = ASTAlloc2(vardefs, $1->line, $1->col, NULL, 0, $1, $2);
          }
          else
          {
             $$ = $2;
          }}
;


stmts: /* empty */ {$$=NULL;}
     | stmts stmt {
          if($1)
          {
             $$ = ASTAlloc2(stmts, $1->line, $1->col, NULL, 0, $1, $2);
          }
          else
          {
             $$ = $2;
          }}
;


stmt: while_stmt   {$$=$1;}
    | if_stmt      {$$=$1;}
    | ret_stmt     {$$=$1;}
    | print_stmt   {$$=$1;}
    | assign       {$$=$1;}
    | funcall_stmt {$$=$1;}
;


while_stmt: WHILE OPENPAR boolexpr CLOSEPAR body {
   $$=ASTAlloc2(while_stmt, $1->line, $1->col, NULL, 0, $3, $5);
 }
;

if_stmt: IF OPENPAR boolexpr CLOSEPAR body{
   $$=ASTAlloc2(if_stmt, $1->line, $1->col, NULL, 0, $3, $5);
 }

;

ret_stmt: RETURN expr SEMICOLON {
   $$=ASTAlloc2(ret_stmt, $1->line, $1->col, NULL, 0, $2, NULL);
 }
;

 print_stmt: PRINT expr SEMICOLON {
    $$=ASTAlloc2(print_stmt, $1->line, $1->col, NULL, 0, $2, NULL);
 }
;


assign: IDENT EQ expr SEMICOLON{
   $$=ASTAlloc2(assign, $1->line, $1->col, NULL, 0, $1, $3);
 }
;

funcall_stmt: funcall SEMICOLON{
   $$ = $1;
 }
;

boolexpr: expr EQ expr  {$$=ASTAlloc2(t_EQ, $1->line, $1->col, NULL, 0, $1, $3);}
        | expr NEQ expr {$$=ASTAlloc2(t_NEQ, $1->line, $1->col, NULL, 0, $1, $3);}
        | expr LT expr  {$$=ASTAlloc2(t_LT, $1->line, $1->col, NULL, 0, $1, $3);}
        | expr GT expr  {$$=ASTAlloc2(t_GT, $1->line, $1->col, NULL, 0, $1, $3);} 
        | expr LEQ expr {$$=ASTAlloc2(t_LEQ, $1->line, $1->col, NULL, 0, $1, $3);}
        | expr GEQ expr {$$=ASTAlloc2(t_GEQ, $1->line, $1->col, NULL, 0, $1, $3);}
;

expr: funcall {$$ = $1;}
    | INTLIT  {$$ = $1;}
    | IDENT   {$$ = $1;}
    | STRINGLIT {$$ = $1;}
    | OPENPAR expr CLOSEPAR {$$ = $2;$$->line=$1->line; $$->col=$1->col;}
    | expr PLUS expr  {$$=ASTAlloc2(t_PLUS, $1->line, $1->col, NULL, 0, $1, $3);}
    | expr MINUS expr {$$=ASTAlloc2(t_MINUS, $1->line, $1->col, NULL, 0, $1, $3);}
    | expr MULT expr  {$$=ASTAlloc2(t_MULT, $1->line, $1->col, NULL, 0, $1, $3);}
    | expr DIV expr   {$$=ASTAlloc2(t_DIV, $1->line, $1->col, NULL, 0, $1, $3);}
    | MINUS expr %prec UMINUS {$$=ASTAlloc2(t_MINUS, $1->line, $1->col, NULL, 0, $2, NULL);} 
;

funcall: IDENT OPENPAR args CLOSEPAR {
   $$=ASTAlloc2(funcall, $1->line, $1->col, NULL, 0, $1, $3);
 }
;

args: /* empty */ {$$=NULL;}
    | arglist {
       $$ = $1;
      }
;

arglist: expr {$$ = $1;}
       |expr COMA arglist {
          $$=ASTAlloc2(arglist, $1->line, $1->col, NULL, 0, $1, $3);
 }
;



%%


int yyerror(const char* err)
{
   printf("%4d:%d: %s\n", yylloc.first_line, yylloc.first_column, err);
   return 0;
}


int main (int argc, char* argv[])
{
  int i;
  int res;

   ++argv, --argc;  /* skip over program name */
   if ( argc > 0 )
   {
      yyin = fopen( argv[0], "r" );
      if(!yyin)
      {
         fprintf(stderr, "Could not open file '%s'\n", argv[0]);
         return EXIT_FAILURE;
      }
   }
   else
   {
      yyin = stdin;
   }   
   res = yyparse();

   if(res==0)
   {
      DOTASTPrint(stdout, ast);
   }
   return res;
}
